package com.audio.device.manager;

import android.Manifest;
import android.app.AlertDialog;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.IntentFilter;
import android.content.ServiceConnection;
import android.content.pm.PackageManager;
import android.graphics.Color;
import android.media.AudioDeviceInfo;
import android.media.AudioManager;
import android.media.AudioRecord;
import android.media.AudioTrack;
import android.os.Build;
import android.os.IBinder;
import android.util.Log;

import androidx.annotation.Nullable;
import androidx.core.app.ActivityCompat;
import com.facebook.react.bridge.Arguments;
import com.facebook.react.bridge.Promise;
import com.facebook.react.bridge.ReactApplicationContext;
import com.facebook.react.bridge.ReactContext;
import com.facebook.react.bridge.ReactContextBaseJavaModule;
import com.facebook.react.bridge.ReactMethod;
import com.facebook.react.bridge.WritableMap;
import com.facebook.react.modules.core.DeviceEventManagerModule;

import java.util.HashSet;
import java.util.Set;

import static android.content.Context.BIND_AUTO_CREATE;
import static com.facebook.react.bridge.UiThreadUtil.runOnUiThread;

public class RNAudioDeviceModule  extends ReactContextBaseJavaModule implements ServiceConnection {
  private static final String REACT_NATIVE_MODULE_NAME = "AudioDeviceManager";
  private static final String TAG = REACT_NATIVE_MODULE_NAME;
  private String mPackageName = "com.audio.device.manager";
  private static final int REQUEST_RECORD_AUDIO_PERMISSION = 1;

  private final ReactApplicationContext reactContext;

  // --- AudioRouteManager
  private AudioManager audioManager;

  // Default audio device; speaker phone for video calls or earpiece for audio
  // only calls.
  private AudioDevice defaultAudioDevice = AudioDevice.NONE;

  private BroadcastReceiver wiredHeadsetReceiver;
  /**
   * AudioDevice is the names of possible audio devices that we currently
   * support.
   */
  public enum AudioDevice {
    SPEAKER_PHONE,
    WIRED_HEADSET,
    EARPIECE,
    BLUETOOTH,
    NONE,
  }

  private boolean hasWiredHeadset = false;

  // Contains the currently selected audio device.
  // This device is changed automatically using a certain scheme where e.g.
  // a wired headset "wins" over speaker phone. It is also possible for a
  // user to explicitly select a device (and overrid any predefined scheme).
  // See |userSelectedAudioDevice| for details.
  private AudioDevice selectedAudioDevice;

  // Contains the user-selected audio device which overrides the predefined
  // selection scheme.
  // TODO(henrika): always set to AudioDevice.NONE today. Add support for
  // explicit selection based on choice by userSelectedAudioDevice.
  private AudioDevice userSelectedAudioDevice;
  private Set<AudioDevice> audioDevices = new HashSet<>();

  // Handles all tasks related to Bluetooth headset devices.
  private final BluetoothManager bluetoothManager;
  private CloudSpeechService mCloudSpeechService;
  private VoiceRecorder mVoiceRecorder;

  public RNAudioDeviceModule(ReactApplicationContext reactContext) {
    super(reactContext);
    this.reactContext = reactContext;

    audioManager =
            (
                    (AudioManager) this.reactContext.getSystemService(Context.AUDIO_SERVICE)
            );

    bluetoothManager = BluetoothManager.create(reactContext, this);
  }

  @Override
  public String getName() {
    return "RNAudioDevice";
  }


  @ReactMethod
  public void start() {
    try {
      Log.d(TAG, "start audioRouteManager");
      bluetoothManager.start();
      setSpeakerphoneOn(true);
      setMicrophoneMute(false);
      hasWiredHeadset = hasWiredHeadset();
      defaultAudioDevice = AudioDevice.SPEAKER_PHONE;
      selectedAudioDevice = AudioDevice.NONE;
      audioDevices.clear();

      Intent intent = new Intent(this.reactContext, CloudSpeechService.class);
      this.reactContext.startService(intent);
      this.reactContext.bindService(intent, this, BIND_AUTO_CREATE);

      if (ActivityCompat.shouldShowRequestPermissionRationale(this.reactContext.getCurrentActivity(),
              Manifest.permission.RECORD_AUDIO)) {
        showPermissionMessageDialog();
      } else {
        ActivityCompat.requestPermissions(this.reactContext.getCurrentActivity(), new String[]{Manifest.permission.RECORD_AUDIO},
                REQUEST_RECORD_AUDIO_PERMISSION);
      }
    } catch (Exception e) {
      Log.e("ERROR", e.getMessage());
    }
  }

  @ReactMethod
  public void stop() {
    Log.d(TAG, "stop() AudioDeviceManager");
    setSpeakerphoneOn(false);
    setMicrophoneMute(false);
    bluetoothManager.stop();

    // Stop Cloud Speech API
    if (mCloudSpeechService != null) {
      mCloudSpeechService.removeListener(mCloudSpeechServiceListener);
      mCloudSpeechService = null;
    }
  }

  @ReactMethod
  public void startRecognize(boolean isRecordable) {
    Log.d(TAG, "Check is recordable voice: " + (isRecordable?"YES":"NO"));

    startVoiceRecorder();
  }

  @ReactMethod
  public void stopRecognize() {
    stopVoiceRecorder();
  }

  private void startVoiceRecorder() {
    if (mVoiceRecorder != null) {
      mVoiceRecorder.stop();
    }

    mVoiceRecorder = new VoiceRecorder(mVoiceCallback, this.reactContext);

    mVoiceRecorder.start();
  }

  private void stopVoiceRecorder() {
    Log.d(TAG, "# stopVoiceRecorder #");
    if (mVoiceRecorder != null) {
      mVoiceRecorder.stop();
      mVoiceRecorder = null;
    }
  }

  private void showPermissionMessageDialog() {
    AlertDialog.Builder alertDialogBuilder = new AlertDialog.Builder(this.reactContext);
    alertDialogBuilder.setMessage("This app needs to record audio and recognize your speech")
            .setPositiveButton(android.R.string.ok, new DialogInterface.OnClickListener() {
              @Override
              public void onClick(DialogInterface dialog, int which) {
              }
            }).setOnDismissListener(new DialogInterface.OnDismissListener() {
      @Override
      public void onDismiss(DialogInterface dialogInterface) {
      }
    }).create();

    AlertDialog alertDialog = alertDialogBuilder.create();

    // show it
    alertDialog.show();
  }

  private final CloudSpeechService.Listener mCloudSpeechServiceListener = new CloudSpeechService.Listener() {
    @Override
    public void onSpeechRecognized(final String text, final boolean isFinal) {
      if (isFinal) {
        mVoiceRecorder.dismiss();
      }

      runOnUiThread(new Runnable() {
        @Override
        public void run() {
          Log.d(TAG, "Final Response : " + text);
          WritableMap data = Arguments.createMap();
          data.putString("text", text);

          sendEvent("onSpeechRecognition", data);
        }
      });
    }
  };


  @Override
  public void onServiceConnected(ComponentName name, IBinder service) {
    mCloudSpeechService = CloudSpeechService.from(service);
    mCloudSpeechService.addListener(mCloudSpeechServiceListener);
  }

  @Override
  public void onServiceDisconnected(ComponentName name) {
    mCloudSpeechService = null;
  }

  private final VoiceRecorder.Callback mVoiceCallback = new VoiceRecorder.Callback() {

    @Override
    public void onVoiceStart() {
      if (mCloudSpeechService != null) {
        Log.d(TAG, "Speech to text start");
        mCloudSpeechService.startRecognizing(mVoiceRecorder.getSampleRate());
      }
    }

    @Override
    public void onVoice(final byte[] buffer, int size) {
      if (mCloudSpeechService != null) {
        Log.d(TAG, "Speech to text progressing");
        mCloudSpeechService.recognize(buffer, size);
      }
    }

    @Override
    public void onVoiceEnd() {
      if (mCloudSpeechService != null) {
        mCloudSpeechService.finishRecognizing();
      }
    }
  };

  /** Helper method for unregistration of an existing receiver. */
  private void unregisterReceiver(final BroadcastReceiver receiver) {
    final ReactContext reactContext = this.getReactApplicationContext();
    if (reactContext != null) {
      try {
        reactContext.unregisterReceiver(receiver);
      } catch (final Exception e) {
        Log.d(TAG, "unregisterReceiver() failed");
      }
    } else {
      Log.d(TAG, "unregisterReceiver() reactContext is null");
    }
  }

  private void setMicrophoneMute(final boolean enable) {
    if (enable != audioManager.isMicrophoneMute()) {
      Log.d(TAG, "setMicrophoneMute(): " + enable);
      audioManager.setMicrophoneMute(enable);
    }
  }

  /**
   * Updates list of possible audio devices and make new device selection.
   */
  public void updateAudioDeviceState() {
    Log.d(
            TAG,
            "--- updateAudioDeviceState: " +
                    "wired headset=" +
                    hasWiredHeadset +
                    ", " +
                    "BT state=" +
                    bluetoothManager.getState()
    );
    Log.d(
            TAG,
            "Device status: " +
                    "available=" +
                    audioDevices +
                    ", " +
                    "selected=" +
                    selectedAudioDevice +
                    ", " +
                    "user selected=" +
                    userSelectedAudioDevice
    );

    // Check if any Bluetooth headset is connected. The internal BT state will
    // change accordingly.
    // TODO(henrika): perhaps wrap required state into BT manager.
    if (
            bluetoothManager.getState() == BluetoothManager.State.HEADSET_AVAILABLE ||
                    bluetoothManager.getState() ==
                            BluetoothManager.State.HEADSET_UNAVAILABLE ||
                    bluetoothManager.getState() == BluetoothManager.State.SCO_DISCONNECTING
    ) {
      bluetoothManager.updateDevice();
    }

    // Update the set of available audio devices.
    Set<AudioDevice> newAudioDevices = new HashSet<>();

    // always assume device has speaker phone
    newAudioDevices.add(AudioDevice.SPEAKER_PHONE);

    if (
            bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTED ||
                    bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTING ||
                    bluetoothManager.getState() == BluetoothManager.State.HEADSET_AVAILABLE
    ) {
      newAudioDevices.add(AudioDevice.BLUETOOTH);
    }

    if (hasWiredHeadset) {
      newAudioDevices.add(AudioDevice.WIRED_HEADSET);
    }

    if (hasEarpiece()) {
      newAudioDevices.add(AudioDevice.EARPIECE);
    }

    // --- check whether user selected audio device is available
    if (
            userSelectedAudioDevice != null &&
                    userSelectedAudioDevice != AudioDevice.NONE &&
                    !newAudioDevices.contains(userSelectedAudioDevice)
    ) {
      userSelectedAudioDevice = AudioDevice.NONE;
    }

    // Store state which is set to true if the device list has changed.
    boolean audioDeviceSetUpdated = !audioDevices.equals(newAudioDevices);
    // Update the existing audio device set.
    audioDevices = newAudioDevices;

    AudioDevice newAudioDevice = getPreferredAudioDevice();

    // --- stop bluetooth if needed
    if (
            selectedAudioDevice == AudioDevice.BLUETOOTH &&
                    newAudioDevice != AudioDevice.BLUETOOTH &&
                    (
                            bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTED ||
                                    bluetoothManager.getState() == BluetoothManager.State.SCO_CONNECTING
                    )
    ) {
      bluetoothManager.stopScoAudio();
      bluetoothManager.updateDevice();
    }

    // --- start bluetooth if needed
    if (
            selectedAudioDevice != AudioDevice.BLUETOOTH &&
                    newAudioDevice == AudioDevice.BLUETOOTH &&
                    bluetoothManager.getState() == BluetoothManager.State.HEADSET_AVAILABLE
    ) {
      // Attempt to start Bluetooth SCO audio (takes a few second to start).
      if (!bluetoothManager.startScoAudio()) {
        // Remove BLUETOOTH from list of available devices since SCO failed.
        audioDevices.remove(AudioDevice.BLUETOOTH);
        audioDeviceSetUpdated = true;
        if (userSelectedAudioDevice == AudioDevice.BLUETOOTH) {
          userSelectedAudioDevice = AudioDevice.NONE;
        }
        newAudioDevice = getPreferredAudioDevice();
      }
    }

    if (
            newAudioDevice == AudioDevice.BLUETOOTH &&
                    bluetoothManager.getState() != BluetoothManager.State.SCO_CONNECTED
    ) {
      newAudioDevice = getPreferredAudioDevice();
    }

    // Switch to new device but only if there has been any changes.
    if (newAudioDevice != selectedAudioDevice || audioDeviceSetUpdated) {
      // Do the required device switch.
      setAudioDeviceInternal(newAudioDevice);
      Log.d(
              TAG,
              "New device status: " +
                      "available=" +
                      audioDevices +
                      ", " +
                      "selected=" +
                      newAudioDevice
      );
      /*
            if (audioManagerEvents != null) {
                // Notify a listening client that audio device has been changed.
                audioManagerEvents.onAudioDeviceChanged(selectedAudioDevice, audioDevices);
            }
            */
      sendEvent("onAudioDeviceChanged", getAudioDeviceStatusMap());
    }
    Log.d(TAG, "--- updateAudioDeviceState done");
  }

  private AudioDevice getPreferredAudioDevice() {
    final AudioDevice newAudioDevice;

    if (
            userSelectedAudioDevice != null &&
                    userSelectedAudioDevice != AudioDevice.NONE
    ) {
      newAudioDevice = userSelectedAudioDevice;
    } else if (audioDevices.contains(AudioDevice.BLUETOOTH)) {
      // If a Bluetooth is connected, then it should be used as output audio
      // device. Note that it is not sufficient that a headset is available;
      // an active SCO channel must also be up and running.
      newAudioDevice = AudioDevice.BLUETOOTH;
    } else if (audioDevices.contains(AudioDevice.WIRED_HEADSET)) {
      // If a wired headset is connected, but Bluetooth is not, then wired headset is used as
      // audio device.
      newAudioDevice = AudioDevice.WIRED_HEADSET;
    } else if (audioDevices.contains(defaultAudioDevice)) {
      newAudioDevice = defaultAudioDevice;
    } else {
      newAudioDevice = AudioDevice.SPEAKER_PHONE;
    }

    return newAudioDevice;
  }

  /** Changes selection of the currently active audio device. */
  private void setAudioDeviceInternal(AudioDevice device) {
    Log.d(TAG, "setAudioDeviceInternal(device=" + device + ")");
    if (!audioDevices.contains(device)) {
      Log.e(TAG, "specified audio device does not exist");
      return;
    }

    switch (device) {
      case SPEAKER_PHONE:
        setSpeakerphoneOn(true);
        break;
      case EARPIECE:
        setSpeakerphoneOn(false);
        break;
      case WIRED_HEADSET:
        setSpeakerphoneOn(false);
        break;
      case BLUETOOTH:
        setSpeakerphoneOn(false);
        break;
      default:
        Log.e(TAG, "Invalid audio device selection");
        break;
    }
    selectedAudioDevice = device;
  }

  private void setSpeakerphoneOn(final boolean enable) {
    if (enable != audioManager.isSpeakerphoneOn()) {
      Log.d(TAG, "setSpeakerphoneOn(): " + enable);
      audioManager.setSpeakerphoneOn(enable);
    }
  }

  private void sendEvent(final String eventName, @Nullable WritableMap params) {
    try {
      ReactContext reactContext = getReactApplicationContext();
      if (reactContext != null && reactContext.hasActiveCatalystInstance()) {
        reactContext
                .getJSModule(DeviceEventManagerModule.RCTDeviceEventEmitter.class)
                .emit(eventName, params);
      } else {
        Log.e(
                TAG,
                "sendEvent(): reactContext is null or not having CatalystInstance yet."
        );
      }
    } catch (RuntimeException e) {
      Log.e(
              TAG,
              "sendEvent(): java.lang.RuntimeException: Trying to invoke JS before CatalystInstance has been set!"
      );
    }
  }

  /** Gets the current earpiece state. */
  private boolean hasEarpiece() {
    return getReactApplicationContext()
            .getPackageManager()
            .hasSystemFeature(PackageManager.FEATURE_TELEPHONY);
  }

  private WritableMap getAudioDeviceStatusMap() {
    WritableMap data = Arguments.createMap();
    String audioDevicesJson = "[";
    for (AudioDevice s : audioDevices) {
      audioDevicesJson += "\"" + s.name() + "\",";
    }

    // --- strip the last `,`
    if (audioDevicesJson.length() > 1) {
      audioDevicesJson =
              audioDevicesJson.substring(0, audioDevicesJson.length() - 1);
    }
    audioDevicesJson += "]";

    data.putString("availableAudioDeviceList", audioDevicesJson);
    data.putString(
            "selectedAudioDevice",
            (selectedAudioDevice == null) ? "" : selectedAudioDevice.name()
    );

    return data;
  }

  /**
   * Checks whether a wired headset is connected or not.
   * This is not a valid indication that audio playback is actually over
   * the wired headset as audio routing depends on other conditions. We
   * only use it as an early indicator (during initialization) of an attached
   * wired headset.
   */
  @Deprecated
  private boolean hasWiredHeadset() {
    if (Build.VERSION.SDK_INT < Build.VERSION_CODES.M) {
      return audioManager.isWiredHeadsetOn();
    } else {
      final AudioDeviceInfo[] devices = audioManager.getDevices(
              AudioManager.GET_DEVICES_ALL
      );
      for (AudioDeviceInfo device : devices) {
        final int type = device.getType();
        if (type == AudioDeviceInfo.TYPE_WIRED_HEADSET) {
          Log.d(TAG, "hasWiredHeadset: found wired headset");
          return true;
        } else if (type == AudioDeviceInfo.TYPE_USB_DEVICE) {
          Log.d(TAG, "hasWiredHeadset: found USB audio device");
          return true;
        }
      }
      return false;
    }
  }
}
